## 数组消除优化

该优化通过计算活跃load、store、memfuncs，得到到达每个指令的store。

通过判断到达当前load的store是否匹配该load且覆盖了其他的所有load，从而执行load替换。

同样，对于不能判断的load，可以对store计算一个冲突范围，对范围内的store增加使用次数。
对于使用次数为0的store，可以认为是没用的store。

对于如下的用例
```cpp
int x[100];
int main(){
    x[1]=10;
    if(getint()>0){
        putint(x[5]);
        putint(x[1]);
        while(xxx){
            putint(x[2]);
        }
    }
    while(xxx){
        int index = getint();
        x[index] = 20;
        while(xxx){
            putint(x[index]);
        }
        if(cond){
            x[1] = 10;
            putint(x[1]);
        }
        putint(x[index]);
    }
}
```
可以优化为
```cpp
int x[100];
int main(){
    x[1]=10;
    if(getint()>0){
        putint(0); //初始化为0，替换
        putint(10);//仅有一个store到达，替换
        while(xxx){
            putint(0);//x1到达域包含此处，且支配，可以替换
        }
    }
    while(xxx){
        int index = getint();
        x[index] = 20;
        while(xxx){
            putint(20);//x[index]覆盖其他store，可以替换
        }
        if(cond){
            x[1] = 10;
            putint(10);//x[1]覆盖其他store，可以替换
        }
        putint(x[index]);//x[1]和x[index]两处store可能冲突，无法替换
    }
}
```

## call函数优化

我们从我们的编译结果中，发现了一种特别的IR模式
```
        //block14:
        //	%16 = icmp slt i32 %num_param, 0
        //	br i32 %16, label %block9, label %block10
        //block9:
        //	ret float 0.0
        //block10:
        //	%20 = sub i32 %num_param, 1
        //	%24 = call float @func(float %data_param, i32 %20)
        //	%25 = fadd float %data_param, %24
        //	%29 = call float @func(float %25, i32 %20)
        //	%30 = fsub float %25, %29
        //	ret float %30
```

这种模式对应的函数判断一个参数（不妨设为参数a，顺序无关）是否小于0，小于0则返回0.0
否则将参数a减一，用这个参数递归调用自身，并传入另一个参数（不妨设参数b）。
* 第一次递归调用自身，传入参数a和参数b，结果驾到参数b上。
* 第二次递归调用自身，传入参数a和参数b，结果从b中减去。

我们发现，可以用数学分析快速解决这种模式。

对于参数a，小于0时返回0，等于0时由于下一次调用小于0，则返回参数b。
a等于1时，b分别变为b+b，再变为2b-2b=0，结果返回0。

对此，我们猜测该函数当参数a是奇数时返回0，偶数时返回参数b，并使用数学归纳法证明：

### 数学归纳证明
其中a、b为变量。“b变为x”意味着在后续的证明中，b等于x，直到下一次赋值。

* 假设对于奇数k作为参数a，小于k的调用均符合上述规律，证明k也符合规律 
  * 由于第一次递归调用传入k-1，根据假设返回0。因此参数b没有变化。 
  * 第二次递归调用传入k-1，根据假设返回0，因此参数b没有变化。 
  * 最后返回b，符合规律。
* 假设对于偶数k为参数a，小于k的调用符合规律，证明k也符合规律
  * 由于第一次递归调用传入k-1，根据假设返回参数b，参数b变为2b。
  * 由于第二次递归调用传入k-1，根据假设返回参数b，参数b变为0。
  * 最后返回0，符合规律

综上所述，我们严格证明了我们的猜想。现在，它成为了定理，我们得以有理有据地对我们定义的IR进行优化。

## LoopPeephole优化

该优化的对象是常见的loop求和并取模常数。

```cpp
sum = initsum;
i = 0;
while(i < maxindex){
    sum += x;
    sum %= remconst; //(remconst < 2^16)
}
use of sum...
```

我们可以将其优化为
```cpp
if(i < maxindex){
    sum = (initsum%remconst)+
            ((maxindex-i)%const * x%const)%const;
else sum = initsum;
```


## blockReorder

对于一个支配时，我们按照直接支配关系重新排序块的顺序。

我们的逻辑是一个块的后继被自己直接支配，则说明该后继的任何执行需要经过该块，则应当优先放在该块后面。

更优先考虑的是进入循环及进入循环体的块。我们假设它们进入的概率更大，需要优先放在该块后面。

对于其他被自己支配的块，只需要在优先考虑前序块的排布后，再考虑即可。

## instReorder

这个优化的作用是减少后端寄存器分配的冲突边。

对于每个后端指令，尝试将其向前移动，如果发现移动后的冲突边数量减少，则实际执行该移动。

由于移动可能会产生新的可移动指令，我们用不动点算法重排直到达到不动点。


